import itertools
import math
import os
import time
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from typing import Any, Callable, Iterator

import torch
import torch.distributed as dist
import torch.nn as nn
from colossalai.accelerator import get_accelerator
from colossalai.booster import Booster
from colossalai.shardformer.policies.auto_policy import _fullname
from cornstarch.pipeline_template import PipelineTemplate
from cornstarch.shardformer.policies.auto_policy import get_autopolicy
from cornstarch.shardformer.policies.pipeline_template_policy import (
    PipelineTemplatePolicyBase,
)
from loguru import logger
from torch.distributed import distributed_c10d
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
from torch.utils.data import DataLoader

from oobleck.engine.configuration_engine import ConfigurationEngine
from oobleck.engine.pipeline_instantiator import PipelineInstantiator
from oobleck.engine.plugin import OobleckPlugin
from oobleck.planning.planner import create_pipeline_templates
from oobleck.planning.profiler import ModelProfiler


class ExecutionEngine:
    """A main execution engine using an execution Backend.

    ExecutionEngine does not have a global view of distributed training.

    """

    def __init__(
        self,
        plugin: OobleckPlugin,
        **booster_kwargs,
    ):
        assert os.environ.get("TORCH_NCCL_ASYNC_ERROR_HANDLING", None) == "0", (
            "Please set TORCH_NCCL_ASYNC_ERROR_HANDLING=0. "
            "Otherwise it is not possible to handle NCCL failure."
        )
        assert (
            not dist.is_initialized()
        ), "Distributed environment must not be initialized."
        assert isinstance(
            plugin, OobleckPlugin
        ), "Plugin must be an instance of OobleckPlugin."

        self.plugin = plugin

        self.pipeline_templates: dict[int, PipelineTemplate] | None = None
        self.booster: Booster | None = None
        self.booster_kwargs = booster_kwargs

        self.notification_receiver_thread: Thread | None = None
        self.need_reconfiguration: bool = False

    @property
    def is_master(self) -> bool:
        configuration_engine = ConfigurationEngine.get_instance()
        if configuration_engine is None:
            raise RuntimeError("ConfigurationEngine must be initialized.")
        return configuration_engine.is_master

    def prepare(
        self,
        model: nn.Module,
        optimizer: Optimizer | None = None,
        criterion: Callable | None = None,
        dataloader: DataLoader | None = None,
        lr_scheduler: LRScheduler | None = None,
    ) -> tuple[nn.Module, Optimizer, Callable, DataLoader, LRScheduler]:
        """Initialize pipeline templates and distributed configuration.

        This function automatically initializes torch.distributed,
        create pipeline templates, instantiate pipelines, and then boost the model.
        """

        assert (
            not dist.is_initialized()
        ), "Distributed environment must not be initialized."

        if self.pipeline_templates is not None:
            raise RuntimeError(
                "Pipeline templates are already initialized. "
                "You should not call prepare() more than once."
            )

        configuration_engine = ConfigurationEngine.get_instance()

        profiler = ModelProfiler(
            configuration_engine.tag,
            model_name_or_path=_fullname(model),
            optimizer_class=_fullname(optimizer),
            config=model.config,
            precision=self.plugin.precision,
            tp_size=self.plugin.tp_size,
            base_dir=configuration_engine.base_dir,
        )

        profile_dataloder = DataLoader(
            dataloader.dataset,
            batch_size=self.plugin.microbatch_size,
            collate_fn=dataloader.collate_fn,
        )
        inputs = next(iter(profile_dataloder))
        profiler.init_profile(inputs)

        configuration_engine.init_distributed()
        profile_data = profiler.load_profile(self.plugin.microbatch_size)

        # Calculate the minimum number of nodes required
        memory = torch.cuda.get_device_properties(0).total_memory
        min_num_nodes = max(
            1,
            math.ceil(sum(layer.mem_required for layer in profile_data) / memory),
        )
        max_num_nodes = configuration_engine.world_size // self.plugin.tp_size

        logger.debug("Creating pipeline templates...")
        model_name = _fullname(model)
        pipeline_templates = create_pipeline_templates(
            model_name,
            profile_data,
            list(range(min_num_nodes, max_num_nodes + 1)),
        )

        policy: PipelineTemplatePolicyBase = get_autopolicy(model_name)
        policy.set_model(model)
        for key in list(pipeline_templates.keys()):
            try:
                template = pipeline_templates[key]
                policy.pipeline_template_sanity_check(template)
            except ValueError as e:
                logger.debug(
                    f"Pipeline template {template} failed to pass sanity check and removed: {e}"
                )
                del pipeline_templates[key]

        if not pipeline_templates:
            raise RuntimeError("No pipeline templates created.")

        self.pipeline_templates = pipeline_templates
        logger.debug(f"Pipeline templates: {self.pipeline_templates}")

        pipeline_instantiator = PipelineInstantiator(
            self.pipeline_templates,
            self.plugin.global_batch_size // self.plugin.microbatch_size,
            self.plugin.fault_tolerance_threshold,
        )
        num_instances, num_microbatches = pipeline_instantiator.instantiate(
            len(configuration_engine.dist_info)
        )
        logger.debug(f"Pipeline instances: {num_instances}")
        logger.debug(f"Microbatches: {num_microbatches}")
        self.plugin.set_pipelines(
            list(
                itertools.chain.from_iterable(
                    itertools.repeat(template, num_templates)
                    for template, num_templates in num_instances.items()
                )
            ),
            num_microbatches,
        )
        self.booster = Booster(plugin=self.plugin, **self.booster_kwargs)
        return self.booster.boost(model, optimizer, criterion, dataloader, lr_scheduler)

    def _estimate_max_num_nodes_required(self):
        # TODO: implement it
        pass

    def notification_receive_func(self):
        logger.info("Start failure notification watcher.")
        configuration_engine = ConfigurationEngine.get_instance()
        immediate_restart = configuration_engine.recv_reconfiguration_notification()
        self.need_reconfiguration = True

        if immediate_restart:
            logger.warning(
                "Immediate reconfiguration is needed. Termiating torch.distributed."
            )
            self.on_receive_reconfiguration_notifiation()
        else:
            logger.info(
                "Received reconfiguration request. Will start after the current iteration."
            )

    def on_receive_reconfiguration_notifiation(self):
        """
        A failure event is received from any worker.
        The reconfiguration engine should reconfigure affected pipelines
        using the set of pipeline templates.
        This function is called in such a case.
        """
        logger.warning("Reconfiguration: shutting down all process groups.")

        device = get_accelerator().get_current_device()
        all_pgs = list(distributed_c10d._pg_map.keys())
        all_pgs = [pg._get_backend(device) for pg in all_pgs]
        all_pgs = [
            pg.wrapped_pg
            if isinstance(pg, distributed_c10d._ProcessGroupWrapper)
            else pg
            for pg in all_pgs
        ]

        with ThreadPoolExecutor(max_workers=len(all_pgs)) as executor:
            executor.map(lambda pg: pg._shutdown(), all_pgs)

        try:
            dist.destroy_process_group(dist.GroupMember.WORLD)
        except Exception:
            pass
        finally:
            # If distributed_c10d says GroupMember.WORLD is None
            # To make sure, clean up internal data structure again
            distributed_c10d._update_default_pg(None)
            distributed_c10d._world.pg_map.clear()
            distributed_c10d._world.pg_names.clear()
            distributed_c10d._world.pg_group_ranks.clear()
            distributed_c10d._world.pg_backend_config.clear()
            distributed_c10d._world.pg_to_tag.clear()
            distributed_c10d._world.tags_to_pg.clear()
            distributed_c10d._world.pg_coalesce_state.clear()
            distributed_c10d._world.pg_default_device.clear()
            distributed_c10d._world.group_count = 0

    def execute(
        self,
        dataloader_iterator: Iterator,
        model: nn.Module,
        criterion: Callable,
        optimizer: Optimizer,
        return_loss: bool = True,
        return_outputs: bool = False,
    ) -> dict[str, Any] | None:
        if self.need_reconfiguration:
            self.notification_receiver_thread.join()
            if dist.is_initialized():
                # If torch.distributed is still initialized, this means
                # the configuration was not immediate and should be done now.
                torch.cuda.synchronize()
                self.on_receive_reconfiguration_notifiation()
            return None

        if getattr(dataloader_iterator, "invalidated", False):
            raise RuntimeError(
                "The dataloader iterator has been invalidated. "
                "Please recreate the iterator before resuming training."
            )

        if (
            self.notification_receiver_thread is None
            or not self.notification_receiver_thread.is_alive()
        ):
            self.notification_receiver_thread = Thread(
                target=self.notification_receive_func,
                name="failure_notification_watcher",
                daemon=True,
            )
            self.notification_receiver_thread.start()

        try:
            return self.booster.execute_pipeline(
                dataloader_iterator,
                model,
                criterion,
                optimizer,
                return_loss=return_loss,
                return_outputs=return_outputs,
            )
        except dist.DistError as e:
            logger.warning(
                f"DistError caught: {str(e)}. Immediate reconfiguration is needed."
            )
            # Failure happens and WORLD process group has been destroyed.
            setattr(dataloader_iterator, "invalidated", True)

            return None

    def reconfigure(
        self, model: nn.Module, optimizer: Optimizer, dataloader: DataLoader
    ) -> tuple[nn.Module, Optimizer, DataLoader]:
        logger.info("Waiting for failure notification watcher to finish.")
        self.notification_receiver_thread.join()

        logger.info("Start reconfiguration in 5 seconds...")
        time.sleep(5)

        assert not dist.is_initialized()
        model, optimizer, dataloader, _ = self.plugin.reconfigure(
            self.pipeline_templates, model, optimizer, dataloader
        )

        self.need_reconfiguration = False
        logger.info("Reconfiguration is done.")
        return model, optimizer, dataloader



from collections import defaultdict

import pulp
from cornstarch.pipeline_template import PipelineTemplate
from loguru import logger


class PipelineInstantiator:
    """A class that determines the number of pipelines to be instantiated
    from each pipeline template and the number of microbatches.
    """

    def __init__(
        self,
        pipeline_templates: dict[int, PipelineTemplate],
        global_num_microbatches: int,
        fault_tolerance_threshold: int,
    ):
        self.pipeline_templates = pipeline_templates
        self.global_num_microbatches = global_num_microbatches
        self.fault_tolerance_threshold = fault_tolerance_threshold

    def instantiate(
        self, num_nodes: int
    ) -> tuple[dict[PipelineTemplate, int], dict[PipelineTemplate, int]]:
        """Instantiate pipelines from pipeline templates.
        If there are already instantiated pipelines, this method will reconfigure
        pipelines for a new distributed configuration.

        Args:
            num_nodes (int): The number of nodes in the distributed environment.

        Returns:
            A tuple of two dict[PipelineTemplate, int] objects, where
            1. dict[PipelineTemplate, int]: the number of pipelines to be instantiated
            2. dict[PipelineTemplate, int]: the number of microbatches per pipeline

        Raises:
            RuntimeError:
                1. If no optimal batch distribution found for the given number of nodes
                   and the global batch size.
                2. If failed to find a set of pipeline templates for the given number of nodes.
        """
        instantiations_options = self._enumerate_instantiation_options(num_nodes)

        # Call self._distribute_batch for each element in instantiations_options
        # Should also include None to properly calculate "index" of optimal dist
        batch_distributions = [
            self.distribute_batch(option) for option in instantiations_options
        ]

        if all(dist is None for dist in batch_distributions):
            raise RuntimeError(
                f"Failed to find optimal batch distribution for {num_nodes} nodes."
            )

        str = "Batch distributions===============\n"
        for latency_dist in batch_distributions:
            if latency_dist is None:
                continue
            latency, dist = latency_dist
            str += f"  {dist} (latency {latency} ms)\n"
        str + "=================================="
        logger.debug(str)

        # Find the second dictionary where its corresponding float is the minimum
        optimal_distribution = min(
            [dist for dist in batch_distributions if dist is not None],
            key=lambda x: x[0],
        )
        index = batch_distributions.index(optimal_distribution)

        logger.info(f"Optimal batch distribution: {optimal_distribution[1]}")

        return (
            instantiations_options[index],
            optimal_distribution[1],
        )

    def _enumerate_instantiation_options(
        self, num_nodes: int
    ) -> list[dict[PipelineTemplate, int]]:
        """Use dynamic programming to get all feasible sets of pipeline templates
        given number of nodes.
        Implementation of Section 4.2.1.

        Args:
            num_nodes (int): The number of nodes in the distributed environment.

        Returns:
            A list of dict[PipelineTemplate, int] objects, where each dict represents
            a feasible set of pipeline templates.
        """
        logger.debug(
            f"Enumerating all feasible sets of pipeline templates for {num_nodes} nodes."
        )

        pipeline_templates: list[PipelineTemplate] = list(
            self.pipeline_templates.values()
        )

        dp: list[list[list[dict[PipelineTemplate, int]]]] = [
            [[] for _ in range(num_nodes + 1)]
            for _ in range(len(self.pipeline_templates) + 1)
        ]

        for i in range(1, len(pipeline_templates) + 1):
            dp[i][0] = [defaultdict(int)]
            for j in range(1, num_nodes + 1):
                # (1) in Figure: copy all dicts
                dp[i][j] = [combo.copy() for combo in dp[i - 1][j]]
                if pipeline_templates[i - 1].num_stages <= j:
                    # (2) in Figure: copy all dicts with one pipeline_templates[i - 1] added
                    for combo in dp[i][j - pipeline_templates[i - 1].num_stages]:
                        new_combo = combo.copy()
                        new_combo[pipeline_templates[i - 1]] += 1

                        # concatenate two lists
                        dp[i][j].append(new_combo)

        logger.debug(f"Dynamic programming result: {dp[-1][-1]}")
        if not dp[-1][-1]:
            raise RuntimeError(
                f"Failed to find feasible sets of pipeline templates for {num_nodes} nodes."
            )
        return [
            option
            for option in dp[-1][-1]
            if sum(option.values()) >= self.fault_tolerance_threshold
        ]

    def distribute_batch(
        self,
        num_pipelines: dict[PipelineTemplate, int],
        need_all_pipelines_have_batch: bool = False,
    ) -> tuple[float, dict[PipelineTemplate, int]] | None:
        """Find the optimal distribution of microbatches that minimizes iteration time.
        Implementation of Section 4.2.2.

        Use integer linear programming library (PuLP) that finds the optimal solution:
        - Let Z the slowest iteration time (max(Bi * Ti))
        - Minimize Z, while keeping sum(Bi) remain constant to global_num_microbatch

        Args:
            global_num_microbatch (int): The total number of microbatches.
            num_pipelines (dict[PipelineTemplate, int]): A set of pipeline templates,
                where each value represents the number of pipelines to be instantiated.

        Returns:
            A tuple of two objects:
                1. float: The optimal iteration time.
                2. A dict[PipelineTemplate, int] object, where each key represents
                   a pipeline template and each value represents the number of microbatches
                   for that pipeline template.

            Or None if the optimal solution is not found.
        """

        assert sum(num_pipelines.values()) >= self.fault_tolerance_threshold, (
            f"The number of pipelines {sum(num_pipelines.values())} should be "
            "greater than or equal to the fault tolerance threshold "
            f"{self.fault_tolerance_threshold}."
        )

        model = pulp.LpProblem("Microbatch Distribution", pulp.LpMinimize)

        # define variables
        num_microbatches = pulp.LpVariable.dicts(
            "num_microbatches",
            num_pipelines.keys(),
            lowBound=1 if need_all_pipelines_have_batch else 0,
            cat=pulp.LpInteger,
        )
        global_iteration_time = pulp.LpVariable("Z", lowBound=0, cat=pulp.LpContinuous)

        # define constraints
        model += (
            sum(
                num_microbatches[template] * num_pipelines[template]
                for template in num_microbatches
            )
            == self.global_num_microbatches
        )
        for template in num_microbatches.keys():
            model += global_iteration_time >= template.latency(
                num_microbatches[template]
            )

        # define objective function
        model += global_iteration_time
        model.solve(pulp.PULP_CBC_CMD(msg=False))

        if pulp.LpStatus[model.status] != "Optimal":
            logger.warning(f"Failed to find optimal solution for {num_pipelines}.")
            return None

        assert (
            float(global_iteration_time.value()) is not None
        ), "Status is optimal but global_iteration_time is None."

        num_microbatches = {
            template: int(num_microbatch.value())
            for template, num_microbatch in num_microbatches.items()
        }
        logger.debug(
            f"Optiomal batch distribution for {num_pipelines}: {num_microbatches}"
        )
        return (float(global_iteration_time.value()), num_microbatches)



from __future__ import annotations

import datetime
import itertools
import os
from multiprocessing.connection import Connection
from pathlib import Path

import torch.distributed as dist
from loguru import logger

from oobleck.elastic.run import HostInfo, HostStatus


class ConfigurationEngine:
    """
    An engine that manages internal configurations and distributed of Oobleck.
    This is the only engine in a worker process that communicates with an agent process.
    Communication with the agent is done via a pipe.
    Users should not touch this class manually.
    """

    _instance: ConfigurationEngine = None

    pipe: Connection
    agent_index: int
    tag: str
    base_dir: Path
    local_rank: int
    dist_info: list[HostInfo]
    rank_map: dict[HostInfo, list[int]]
    rank: int

    def __init__(self):
        raise NotImplementedError(
            "Use get_instance() instead to get an instance of ConfigurationEngine."
        )

    @staticmethod
    def create(
        pipe: Connection,
        agent_index: int,
        local_rank: int,
        tag: str,
        base_dir: Path,
    ) -> ConfigurationEngine:
        """Create a new instance of ConfigurationEngine."""
        if ConfigurationEngine._instance is not None:
            return ConfigurationEngine._instance

        instance = ConfigurationEngine.__new__(ConfigurationEngine)

        # TODO: set initial attributes.
        instance.pipe = pipe
        instance.agent_index = agent_index
        instance.tag = tag
        instance.base_dir = base_dir

        instance.local_rank = local_rank
        dist_info: list[HostInfo] = pipe.recv()
        instance.dist_info = dist_info

        logger.debug(f"dist_info: {dist_info}, agent_index: {agent_index}")

        instance.rank_map = {
            host: list(range(i * len(gpu_indices), (i + 1) * len(gpu_indices)))
            for i, host in enumerate(dist_info)
            if (gpu_indices := host.devices.split(","))
        }
        my_agent = dist_info[agent_index]
        instance.rank = instance.rank_map[my_agent][instance.local_rank]

        logger.debug(f"rank_map: {instance.rank_map}")

        ConfigurationEngine._instance = instance
        return ConfigurationEngine._instance

    @staticmethod
    def get_instance() -> ConfigurationEngine:
        assert (
            ConfigurationEngine._instance is not None
        ), "ConfigurationEngine is not initialized."
        return ConfigurationEngine._instance

    def get_host_update(self):
        """
        Get host update from the agent process.
        """
        my_agent = self.dist_info[self.agent_index]
        new_dist_info: list[HostInfo] = self.pipe.recv()

        assert not any(
            host.status == HostStatus.killed for host in new_dist_info
        ), "Worker should not see any killed agent status."

        agent_index = new_dist_info.index(my_agent)
        if new_dist_info[agent_index].status == HostStatus.terminating:
            # This process will be terminated after the iteration.
            logger.info(
                f"Worker {self.local_rank} in agent {self.agent_index} terminating..."
            )
            os._exit(0)

        self.dist_info = [
            host_info
            for host_info in new_dist_info
            if host_info.status == HostStatus.up
        ]
        self.agent_index = self.dist_info.index(my_agent)

        self.rank_map = {
            host: list(range(i * len(gpu_indices), (i + 1) * len(gpu_indices)))
            for i, host in enumerate(self.dist_info)
            if (gpu_indices := host.devices.split(","))
        }
        my_agent = self.dist_info[self.agent_index]
        self.rank = self.rank_map[my_agent][self.local_rank]

    @property
    def all_ranks(self) -> list[int]:
        return itertools.chain.from_iterable(self.rank_map.values())

    @property
    def world_size(self) -> int:
        if not self.dist_info:
            return 0

        num_devices = len(self.dist_info[0].devices.split(","))
        return num_devices * len(self.dist_info)

    @property
    def is_master(self) -> bool:
        """Return True if the current process is the master process (rank 0).
        Because it is not guarantted to have torch.distributed initialized,
        it is not possible to use dist.get_rank() == 0 to check.
        Instead, use agent index and local rank which are known without torch.ditributed.
        """
        return self.agent_index == 0 and self.local_rank == 0

    def recv_reconfiguration_notification(self) -> bool:
        """
        Block this thread until the agent sends a reconfiguration message.

        Returns:
            True if training requests immediate reconfiguration.
            False if some agents wll be terminated after the iteration.
        """
        try:
            message = self.pipe.recv()
            if message == "immediate_reconfigure":
                return True
            elif message == "reconfigure":
                return False

            raise ValueError(f"Unexpected reconfiguration message: {message}")
        except Exception:
            # Corresponding agent died. This process should also die.
            os._exit(1)

    def send_distributed_port(self, port: int):
        self.pipe.send(port)

    def receive_distributed_port(self) -> int:
        return self.pipe.recv()

    def init_distributed(self):
        """
        Initialize torch.distributed.

        When it was initialized before, destroy it first and reinitialize.
        When destruction happened, the function returns the old dist_info.
        """
        if dist.is_initialized():
            # TODO: if we try to destroy a process group where some operation is stuck,
            # destroying it might be stuck as well.
            # If this is witnessed, change it to destryoing all process groups
            # manually gathered in ThreadPoolExecutor.
            dist.destroy_process_group(dist.GroupMember.WORLD)

        assert not dist.is_initialized()

        if self.is_master:
            store = dist.TCPStore(
                host_name=self.dist_info[0].ip,
                port=0,
                world_size=self.world_size,
                is_master=True,
                wait_for_workers=False,
            )
            logger.debug(f"torch rank 0 port: {store.port}")
            self.send_distributed_port(store.port)
            # this distributed port is broadcasted and event this process receives it.
            # For master it is useless, so just discard it.
            self.receive_distributed_port()

            # Send anything to the agent back to ask to reset the port
            self.pipe.send(0)
        else:
            port = self.receive_distributed_port()
            logger.debug(f"Received torch.distributed rank 0 port: {port}")
            store = dist.TCPStore(
                host_name=self.dist_info[0].ip,
                port=port,
                world_size=self.world_size,
                is_master=False,
                wait_for_workers=False,
            )

        logger.debug(
            "Initializing torch.distributed. "
            f"rank: {self.rank}, world size: {self.world_size}"
        )

        dist.init_process_group(
            backend="nccl",
            store=store,
            rank=self.rank,
            world_size=self.world_size,
            timeout=datetime.timedelta(minutes=5),
        )

        assert dist.is_initialized(), "Distributed environment is not initialized."
        logger.debug("Distributed environment initialized.")



import multiprocessing
import os
import runpy
import sys
import threading
import time
from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing.connection import Connection
from multiprocessing.context import SpawnContext, SpawnProcess
from pathlib import Path
from typing import cast

import click
import grpc
import torch
from google.protobuf.empty_pb2 import Empty
from loguru import logger

from oobleck.elastic.master_service_pb2 import CodeInfo, DistInfo, PortInfo
from oobleck.elastic.master_service_pb2_grpc import OobleckMasterStub
from oobleck.elastic.run import HostInfo, HostStatus
from oobleck.engine.configuration_engine import ConfigurationEngine


@contextmanager
def temporary_argv(new_argv: list[str]):
    old_argv = sys.argv[:]
    sys.argv = new_argv
    try:
        yield
    finally:
        sys.argv = old_argv


@dataclass
class Worker:
    pipe: Connection
    process: SpawnProcess

    @staticmethod
    def worker_main(
        pipe: Connection,
        agent_index: int,
        gpu_index: int,
        tag: str,
        base_dir: Path,
        script_path: Path,
        script_args: list[str],
    ):
        """
        Worker process main function.

        It creates ConfigurationEngine that will internally be used in
        ExecutionEngine, and execute the given code.
        """
        assert (
            torch.cuda.device_count() == 1
        ), "CUDA_VISIBLE_DEVICES must be set to a single GPU."

        assert ConfigurationEngine._instance is None, (
            "ConfigurationEngine must not be initialized before "
            "worker_main() is called."
        )

        logger.debug(
            f"Worker process started: (agent_index: {agent_index}, "
            f"gpu_index: {os.environ['CUDA_VISIBLE_DEVICES']})"
        )

        ConfigurationEngine.create(pipe, agent_index, gpu_index, tag, base_dir)

        argv = [script_path] + list(script_args)
        with temporary_argv(argv):
            runpy.run_path(script_path.as_posix(), run_name="__main__")


class Agent:
    """Oobleck Agent class.

    For each node, there is one agent process that manages
    worker processes in the node.
    """

    def __init__(
        self,
        agent_index: int,
        job_tag: str,
        base_dir: Path,
        stub: OobleckMasterStub,
    ):
        self.agent_index = agent_index
        self.tag = job_tag
        self.base_dir = base_dir
        self.stub = stub

        # Get distributed information and code from the master
        dist_info: DistInfo = stub.GetDistInfo(Empty())
        self.dist_info = list(
            HostInfo(host.ip, host.devices, host.port, HostStatus[host.status])
            for host in dist_info.hosts
        )
        training_args: CodeInfo = stub.GetCode(Empty())
        self.script: Path = Path(training_args.path)
        self.script_args: list[str] = [arg for arg in training_args.args]
        self.workers: list[Worker] = []

    def notify_reconfiguration_to_workers(
        self, dist_info: list[HostInfo], immediate_restart: bool
    ):
        logger.warning(
            f"Reconfiguration request received from master: {dist_info}. Sending to workers"
        )
        for worker in self.workers:
            worker.pipe.send(
                "immediate_reconfigure" if immediate_restart else "reconfigure"
            )
            worker.pipe.send(dist_info)

        # If this agent is about to die, don't forward the port
        if dist_info[self.agent_index].status == HostStatus.terminating:
            return

        self.forward_master_port()

    def watch_reconfiguration_notification(self):
        for dist_info in self.stub.WatchReconfigurationNotification(Empty()):
            dist_info = cast(DistInfo, dist_info)
            dist_info = [
                HostInfo(host.ip, host.devices, host.port, HostStatus[host.status])
                for host in dist_info.hosts
            ]

            immediate_restart = False
            if any(host.status == HostStatus.killed for host in dist_info):
                immediate_restart = True
            else:
                assert (
                    len(self.dist_info) != len(dist_info)
                    or any(host.status == HostStatus.terminating for host in dist_info)
                ), "The number of hosts must not change or some hosts should be in terminating."

            self.dist_info = [
                host_info
                for host_info in dist_info
                if host_info.status != HostStatus.killed
            ]
            self.notify_reconfiguration_to_workers(self.dist_info, immediate_restart)

    def run_profiler(self):
        raise NotImplementedError()

    def launch_workers(self):
        """Launch worker processes."""
        ctx: SpawnContext = multiprocessing.get_context("spawn")

        gpu_indices: list[int] = list(
            int(dev) for dev in self.dist_info[self.agent_index].devices.split(",")
        )

        tensor_parallel_size = len(gpu_indices)
        ranks = range(
            self.agent_index * tensor_parallel_size,
            (self.agent_index + 1) * tensor_parallel_size,
        )

        os.environ["TORCH_NCCL_USE_COMM_NONBLOCKING"] = "1"
        os.environ["TORCH_NCCL_ASYNC_ERROR_HANDLING"] = "0"
        for gpu_index, rank in zip(gpu_indices, ranks):
            logger.info(f"Launching worker {rank} (GPU: {gpu_index})...")

            pipe, child_pipe = ctx.Pipe()

            os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_index)
            process: SpawnProcess = ctx.Process(
                target=Worker.worker_main,
                args=(
                    child_pipe,
                    self.agent_index,
                    gpu_index % tensor_parallel_size,
                    self.tag,
                    self.base_dir,
                    self.script,
                    self.script_args,
                ),
                daemon=False,
            )
            process.start()
            self.workers.append(Worker(pipe, process))
            pipe.send(self.dist_info)
        del os.environ["CUDA_VISIBLE_DEVICES"]

        self.forward_master_port()

    def forward_master_port(self):
        """
        Forward master port after receiving it from the master
        to all worker processes.

        If this is the first agent, it should send the rank 0 port
        from its first worker to master too.
        """
        # If this is the first agent, it should forward the master rank port
        if self.agent_index == 0:
            logger.debug("Waiting for rank 0 port...")
            port: int = self.workers[0].pipe.recv()
            logger.debug(f"Received rank 0 port: {port}. Sending it to master.")
            self.stub.SetMasterRankPort(PortInfo(port=port))

        port: int = 0
        while port == 0:
            time.sleep(0.1)
            port = self.stub.GetMasterRankPort(Empty()).port

        for worker in self.workers:
            worker.pipe.send(port)

        # Master rank will send another message to the agent to reset the port
        if self.agent_index == 0:
            self.workers[0].pipe.recv()
            self.stub.SetMasterRankPort(PortInfo(port=0))

    def watch_worker_exit(self):
        """Watch worker exit and restart it.
        TODO: It must detect ANY worker exit, not just the first one."""
        for worker in self.workers:
            worker.process.join()
            if worker.process.exitcode != 0:
                logger.warning(
                    f"Worker {worker.process.pid} exited with code {worker.process.exitcode}."
                )
        logger.info("All workers exited.")


@click.command
@click.option("--master_ip", type=str, help="Master IP address.")
@click.option("--master_port", type=int, help="Master port.")
@click.option("--agent_index", type=int, help="The index of this agent process.")
@click.option("--tag", type=str, help="A tag to identify this run.")
@click.option(
    "--base_dir", type=Path, help="Oobleck root directory store logs and profiles."
)
def run(master_ip: str, master_port: int, agent_index: int, tag: str, base_dir: Path):
    # Connect to the master
    channel = grpc.insecure_channel(f"{master_ip}:{master_port}")
    stub = OobleckMasterStub(channel)

    agent = Agent(agent_index, tag, base_dir, stub)
    agent.launch_workers()

    def watch_reconfiguration_noti_func():
        while True:
            agent.watch_reconfiguration_notification()

    thread = threading.Thread(target=watch_reconfiguration_noti_func, daemon=True)
    thread.start()

    agent.watch_worker_exit()
    logger.info(f"Agent {agent_index} exited.")


if __name__ == "__main__":
    run()



import click
import grpc
from google.protobuf.empty_pb2 import Empty

from oobleck.elastic.master_service_pb2 import AgentInfo, DistInfo
from oobleck.elastic.master_service_pb2_grpc import OobleckMasterStub


@click.group(help="CLI tool to query and manipulate Oobleck training.")
@click.option("--ip", type=str, default="localhost", help="Master IP address.")
@click.option("--port", type=int, help="Master port.")
@click.pass_context
def main(ctx: click.core.Context, ip: str, port: int):
    channel = grpc.insecure_channel(f"{ip}:{port}")
    stub = OobleckMasterStub(channel)

    ctx.ensure_object(dict)
    ctx.obj["stub"] = stub


@main.command()
@click.pass_context
def get_agent_list(ctx: click.core.Context):
    stub: OobleckMasterStub = ctx.obj["stub"]
    response: DistInfo = stub.GetDistInfo(Empty())

    print("=== Agents ===")
    for index, host in enumerate(response.hosts):
        print(
            f"[{index}] IP: {host.ip}:{host.port} Status: {host.status} (device indices: {host.devices})"
        )
    print("==============")


@main.command()
@click.pass_context
@click.option("--agent_index", type=int)
def kill_agent(ctx: click.core.Context, agent_index: int):
    stub: OobleckMasterStub = ctx.obj["stub"]
    stub.KillAgent(AgentInfo(agent_index=agent_index))


if __name__ == "__main__":
    main(obj={})



from cornstarch.pipeline_template import PipelineTemplate

from oobleck.planning.profiler import LayerExecutionResult

def create_pipeline_templates(
    model_name: str,
    profile_data: list[LayerExecutionResult],
    num_nodes: list[int],
) -> dict[int, PipelineTemplate]: ...



import functools
import importlib
import json
from dataclasses import asdict, dataclass, field
from enum import Enum
from functools import reduce
from pathlib import Path

import torch
import torch.distributed as dist
import torch.nn as nn
from colossalai.accelerator import get_accelerator
from colossalai.amp.naive_amp.mixed_precision_optimizer import MixedPrecisionOptimizer
from colossalai.booster.plugin.hybrid_parallel_plugin import get_param_info
from colossalai.interface import OptimizerWrapper
from colossalai.shardformer import ShardConfig, ShardFormer
from cornstarch.pipeline_template import PipelineTemplate
from loguru import logger
from torch.distributed import FileStore
from transformers import PretrainedConfig, PreTrainedModel

from oobleck.engine.configuration_engine import ConfigurationEngine


@dataclass
class LayerExecutionResult:
    layer_index: int
    layer_name: str
    forward: float
    backward: float
    mem_required: int


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, LayerExecutionResult):
            return asdict(obj)
        return super().default(obj)


class ModelProfiler:
    """A class for profiling a model.

    Profiling includes:
    - Forward and backward latency (in ms) for each layer
    - Maximum memory consumption (in bytes) for each layer

    Args:
        model (nn.Module): The model to be profiled.
        layers (list[str]): A list of layer names to be profiled.
        model must have modules with the given names.
    """

    def __init__(
        self,
        tag: str,
        model_name_or_path: str,
        optimizer_class: str,
        config: PretrainedConfig,
        precision: str,
        tp_size: int,
        base_dir: Path,
    ):
        self.model_name_or_path = model_name_or_path
        self.optimizer_class = optimizer_class
        self.precision = precision
        self.model_config = config
        self.tp_size = tp_size
        self.profile_dir = base_dir / tag / "profile"

    @staticmethod
    def get_profile_path(
        profile_dir: Path, tp_size: int, microbatch_size: int, precision: str
    ) -> Path:
        profile_dir.mkdir(parents=True, exist_ok=True)
        return profile_dir / f"profile_tp{tp_size}_mb{microbatch_size}_{precision}.json"

    def init_profile(self, inputs: dict[str, torch.Tensor]):
        """Profile the model with a new child process.

        All processes calls this function, and all workers in the first agent
        actually profile the model.
        """
        assert not dist.is_initialized(), "torch.distributed should not be initialized."
        configuration_engine = ConfigurationEngine.get_instance()

        if configuration_engine.agent_index != 0:
            return

        microbatch_size = inputs["input_ids"].shape[0]
        profile_path = ModelProfiler.get_profile_path(
            self.profile_dir, self.tp_size, microbatch_size, self.precision
        )
        if profile_path.exists():
            logger.debug(f"Profile exists: {profile_path}")
            return

        context = torch.multiprocessing.get_context("spawn")
        process = context.Process(
            target=ModelProfiler._profile_model,
            kwargs={
                "model_name_or_path": self.model_name_or_path,
                "model_config": self.model_config,
                "optimizer_class": self.optimizer_class,
                "profile_dir": self.profile_dir,
                "local_rank": configuration_engine.local_rank,
                "tp_size": self.tp_size,
                "precision": self.precision,
                "inputs": inputs,
            },
            daemon=True,
        )
        process.start()
        process.join()

    def load_profile(self, microbatch_size: int) -> list[LayerExecutionResult]:
        """Load profile data from storage.

        Only rank 0 loads profile data, which is broadcasted to all others.
        """
        assert dist.is_initialized(), "torch.distributed is not initialized."

        configuration_engine = ConfigurationEngine.get_instance()
        device = get_accelerator().get_current_device()
        size_tensor = torch.empty(1, dtype=torch.int64, device=device)
        if configuration_engine.rank == 0:
            profile_path = ModelProfiler.get_profile_path(
                self.profile_dir, self.tp_size, microbatch_size, self.precision
            )
            data = profile_path.read_bytes()
            data_tensor = torch.tensor(list(data), dtype=torch.uint8, device=device)
            size_tensor[0] = data_tensor.numel()

        dist.broadcast(size_tensor, src=0)

        if configuration_engine.rank != 0:
            data_tensor = torch.empty(
                size_tensor.item(), dtype=torch.uint8, device=device
            )

        dist.broadcast(data_tensor, src=0)
        torch.cuda.synchronize()

        data = json.loads(data_tensor.cpu().numpy().tobytes())
        return [
            LayerExecutionResult(
                layer_index=layer["layer_index"],
                layer_name=layer["layer_name"],
                forward=layer["forward"],
                backward=layer["backward"],
                mem_required=layer["mem_required"],
            )
            for layer in data["layers"]
        ]

    @staticmethod
    def get_module_by_name(model: nn.Module, name: str) -> nn.Module:
        """Get a module by its name."""
        names = name.split(".")
        return reduce(getattr, names, model)

    @staticmethod
    def _profile_model(
        model_name_or_path: str,
        model_config: PretrainedConfig,
        optimizer_class: str,
        profile_dir: Path,
        local_rank: int,
        tp_size: int,
        precision: str,
        inputs: dict[str, torch.Tensor],
        warmup: int = 3,
    ):
        class EventTiming(Enum):
            FORWARD_START = 0
            FORWARD_END = 1
            BACKWARD_START = 2
            BACKWARD_END = 3
            OPTIMIZER_STEP_START = 4
            OPTIMIZER_STEP_END = 5

        @dataclass
        class ProfileData:
            module_name: str
            events: dict[EventTiming, torch.cuda.Event] = field(default_factory=dict)
            memory: dict[EventTiming, int] = field(default_factory=dict)

        store_path = profile_dir / "store"
        logger.debug(
            f"Profiler initiating torch.distributed: {store_path} with {tp_size} workers"
        )

        store = FileStore(str(store_path), tp_size)
        dist.init_process_group(
            backend="nccl",
            world_size=tp_size,
            rank=local_rank,
            store=store,
        )

        assert dist.get_world_size() == tp_size, "World size mismatch"
        logger.debug(f"Sharding model with {tp_size} ranks")

        module_name, cls = model_name_or_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        model: PreTrainedModel = getattr(module, cls)(model_config).to("cpu")
        model.gradient_checkpointing_enable()

        layers = PipelineTemplate.get_modules(model)
        profile_data: dict[str, ProfileData] = {
            layer_name: ProfileData(
                module_name=layer_name,
                events={
                    timing: torch.cuda.Event(enable_timing=True)
                    for timing in EventTiming
                },
            )
            for layer_name in layers
        }

        optim_name, cls = optimizer_class.rsplit(".", 1)
        module = importlib.import_module(optim_name)
        optim_cls = getattr(module, cls)

        if tp_size > 1:
            # FIXME (insujang): copy user shardconfig as well
            shard_config = ShardConfig(
                tensor_parallel_process_group=dist.new_group(),
                pipeline_stage_manager=None,
                enable_tensor_parallelism=True,
                enable_flash_attention=False,
            )
            shardformer = ShardFormer(shard_config)

            model, _ = shardformer.optimize(model)

        mixed_precision = None
        if precision == "fp16":
            mixed_precision = torch.float16
        elif precision == "bf16":
            mixed_precision = torch.bfloat16
        if mixed_precision is not None:
            model = model.to(dtype=mixed_precision)

        optimizer = optim_cls(model.parameters())
        optim_param_info = get_param_info(optimizer)
        if precision in ["fp16", "bf16"]:
            optimizer = MixedPrecisionOptimizer(optimizer, precision=precision)
        else:
            optimizer = OptimizerWrapper(optimizer)

        # Configure hooks for each layer
        def forward_pre_hook(module_name: str, module: nn.Module, inputs):
            module.to("cuda")
            profile_data[module_name].memory[EventTiming.FORWARD_START] = (
                torch.cuda.memory_allocated()
            )
            event = profile_data[module_name].events[EventTiming.FORWARD_START]
            event.record()

        def forward_hook(module_name: str, module: nn.Module, inputs, outputs):
            profile_data[module_name].memory[EventTiming.FORWARD_END] = (
                torch.cuda.memory_allocated()
            )
            event = profile_data[module_name].events[EventTiming.FORWARD_END]
            event.record()
            module.to("cpu")

        modules_to_offload: list[tuple[str, torch.nn.Module]] = []

        def backward_pre_hook(module_name: str, module: nn.Module, grad_output):
            module.to("cuda")
            profile_data[module_name].memory[EventTiming.BACKWARD_START] = (
                torch.cuda.memory_allocated()
            )
            event = profile_data[module_name].events[EventTiming.BACKWARD_START]
            event.record()

        def backward_hook(module_name, module: nn.Module, grad_input, grad_output):
            event = profile_data[module_name].events[EventTiming.BACKWARD_END]
            event.record()

            profile_data[module_name].memory[EventTiming.BACKWARD_END] = profile_data[
                module_name
            ].memory[EventTiming.BACKWARD_START] + sum(
                p.numel() * p.element_size() for p in module.parameters()
            )

            for _, m in modules_to_offload:
                m.to("cpu")
            modules_to_offload.clear()
            modules_to_offload.append((module_name, module))

        # Move inputs to cuda
        for name in inputs.keys():
            inputs[name] = inputs[name].to("cuda")

        logger.info("Profiler started...")

        for layer_name in layers:
            module = ModelProfiler.get_module_by_name(model, layer_name)

            module.register_forward_pre_hook(
                functools.partial(forward_pre_hook, layer_name)
            )
            module.register_forward_hook(functools.partial(forward_hook, layer_name))
            module.register_full_backward_pre_hook(
                functools.partial(backward_pre_hook, layer_name)
            )
            module.register_full_backward_hook(
                functools.partial(backward_hook, layer_name)
            )

        with torch.no_grad():
            for _ in range(warmup):
                model(**inputs)

        should_continue: bool = True

        while should_continue:
            for param in model.parameters():
                param.grad = None

            logger.debug("Iterating until overflow solved...")
            outputs = model(**inputs)
            optimizer.backward(outputs.loss)
            torch.cuda.synchronize()

            modules_to_offload.clear()
            for param in model.parameters():
                param.data = param.data.to("cpu")
                param.grad.data = param.grad.data.to("cpu")

            if (
                mixed_precision is None
                or not optimizer.mixed_precision.should_skip_step()
            ):
                should_continue = False

                for layer_name in layers:
                    profile_data[layer_name].memory[
                        EventTiming.OPTIMIZER_STEP_START
                    ] = 0
                    profile_data[layer_name].memory[EventTiming.OPTIMIZER_STEP_END] = 0
                optimizer.step()

                num_parameters = 0
                working_to_master_map = (
                    optimizer.get_working_to_master_map()
                    if (precision in ["fp16", "bf16"])
                    else None
                )
                for layer_name in layers:
                    module = ModelProfiler.get_module_by_name(model, layer_name)
                    for param_name, p in module.named_parameters():
                        if f"{layer_name}.{param_name}" in model._tied_weights_keys:
                            continue

                        if precision in ["fp16", "bf16"]:
                            optim_param_index_id = optim_param_info["id2param"][
                                num_parameters
                            ]
                            master_tensor = working_to_master_map[optim_param_index_id]
                            states: dict[torch.Tensor, dict] = (
                                optimizer.optim.state.get(master_tensor)
                            )
                        else:
                            states: dict[torch.Tensor, dict] = (
                                optimizer.optim.state.get(p)
                            )

                        if states:
                            for name, state in states.items():
                                if isinstance(state, torch.Tensor):
                                    profile_data[layer_name].memory[
                                        EventTiming.OPTIMIZER_STEP_END
                                    ] += state.numel() * state.element_size()

                        num_parameters += 1

            optimizer.zero_grad()

        logger.debug("Profiler finished.")

        rank = dist.get_rank()
        if rank == 0:
            microbatch_size = inputs["input_ids"].shape[0]
            profile_path = ModelProfiler.get_profile_path(
                profile_dir, tp_size, microbatch_size, precision
            )
            logger.debug(f"Writing results to {profile_path}")

            data = {
                "model_name": model_name_or_path,
                "microbatch_size": microbatch_size,
                "tp_size": tp_size,
                "precision": precision,
                "layers": [],
            }
            for index, (layer_name, layer_profile) in enumerate(profile_data.items()):
                data["layers"].append(
                    asdict(
                        LayerExecutionResult(
                            layer_index=index,
                            layer_name=layer_name,
                            forward=layer_profile.events[
                                EventTiming.FORWARD_START
                            ].elapsed_time(
                                layer_profile.events[EventTiming.FORWARD_END]
                            ),
                            backward=layer_profile.events[
                                EventTiming.BACKWARD_START
                            ].elapsed_time(
                                layer_profile.events[EventTiming.BACKWARD_END]
                            ),
                            mem_required=(
                                layer_profile.memory[EventTiming.FORWARD_END]
                                - layer_profile.memory[EventTiming.FORWARD_START]
                            )
                            + (
                                layer_profile.memory[EventTiming.BACKWARD_END]
                                - layer_profile.memory[EventTiming.BACKWARD_START]
                            )
                            + (
                                layer_profile.memory[EventTiming.OPTIMIZER_STEP_END]
                                - layer_profile.memory[EventTiming.OPTIMIZER_STEP_START]
                            ),
                        )
                    )
                )

            with profile_path.open("w") as f:
                json.dump(data, f, cls=JsonEncoder)

        dist.barrier()
        torch.cuda.synchronize()

        dist.destroy_process_group()

        if rank == 0:
            store_path.unlink()



import copy
import io
import itertools
from collections import Counter, defaultdict
from typing import Optional

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from colossalai.accelerator import get_accelerator
from colossalai.amp.naive_amp.mixed_precision_optimizer import MixedPrecisionOptimizer
from colossalai.booster.plugin.hybrid_parallel_plugin import (
    TP_AXIS,
    HybridParallelAMPOptimizer,
    HybridParallelNaiveOptimizer,
)
from colossalai.shardformer.layer.parallel_module import ParallelModule
from cornstarch import (
    HeterogeneousDataLoader,
    HeterogeneousParallelModule,
    HeterogeneousParallelPlugin,
    PipelineTemplate,
)
from cornstarch.process_group_mesh import (
    PP_AXIS,
    HeterogeneousProcessGroupMesh,
)
from cornstarch.shardformer.shard.shardformer import ModelSharder
from loguru import logger
from torch.optim.lr_scheduler import LRScheduler

from oobleck.engine.configuration_engine import ConfigurationEngine, HostInfo
from oobleck.engine.pipeline_instantiator import PipelineInstantiator


class OobleckPlugin(HeterogeneousParallelPlugin):
    """Plugin for Oobleck, an extension of heterogeneous parallel plugin
    to support fault tolerance and reconfiguration.
    """

    def __init__(
        self,
        tp_size: int,
        global_batch_size: int,
        microbatch_size: int,
        fault_tolerance_threshold: int = 3,
        precision: str = "fp16",
        enable_fused_normalization: bool = False,
        enable_flash_attention: bool = False,
        enable_jit_used: bool = False,
    ):
        assert (
            global_batch_size % microbatch_size == 0
        ), "Global batch size must be divisible by microbatch size. "

        configuration_engine = ConfigurationEngine.get_instance()
        assert all(
            len(host.devices.split(",")) == tp_size
            for host in configuration_engine.dist_info
        ), f"All agent must have the same worker size {tp_size}."

        super().__init__(
            tp_size=tp_size,
            microbatch_size=microbatch_size,
            precision=precision,
            enable_fused_normalization=enable_fused_normalization,
            enable_flash_attention=enable_flash_attention,
            enable_jit_fused=enable_jit_used,
        )

        self.global_batch_size = global_batch_size
        self.fault_tolerance_threshold = fault_tolerance_threshold

    def _instantiate_pipelines(
        self,
        pipeline_templates: dict[int, PipelineTemplate],
        global_num_microbatches: int,
        old_pg_mesh: Optional[list] = None,
        old_rank_map: Optional[dict[HostInfo, list[int]]] = None,
    ) -> tuple[list[PipelineTemplate], dict[PipelineTemplate, int]]:
        """
        Get pipelines to be instantiated from the list of pipeline templates.
        Each pipeline template may be instantiated several times.

        Args:
            pipeline_templates: Dict of pipeline templates (num_hosts -> pipeline template)
            global_num_microbatches: Number of microbatches in a global batch
            old_pg_mesh: previous group rank mesh (HeterogeneousProcessGroupMesh.mesh).
                Necessary for reconfiguration. None if not reconfiguration.
            old_rank_map: previous rank map (ConfigurationEngine.rank_map).
                Necessary for reconfiguration. None if not reconfiguration.

        Returns:
            list[PipelineTemplate]: List of pipelines to be instantiated
            dict[PipelineTemplate, int]: Number of microbatches for each pipeline
        """
        pipeline_instantiator = PipelineInstantiator(
            pipeline_templates, global_num_microbatches, self.fault_tolerance_threshold
        )
        configuration_engine = ConfigurationEngine.get_instance()

        # Is this for reconfiguration?
        if old_pg_mesh is not None and old_rank_map is not None:
            if not hasattr(self, "pipelines"):
                raise RuntimeError(
                    "Existing pipelines are necessary for reconfiguration"
                )

            num_hosts_per_pipeline = [
                pipeline.num_stages for pipeline in self.pipelines
            ]
            removed_ranks_list = [
                old_rank_map[host]
                for host in old_rank_map
                if host not in configuration_engine.rank_map
            ]

            for removed_ranks in removed_ranks_list:
                for pipeline_index, ranks_in_mesh in enumerate(old_pg_mesh):
                    if removed_ranks in ranks_in_mesh:
                        num_hosts_per_pipeline[pipeline_index] -= 1

            # TODO (insujang): If there is no feasible pipeline, merge pipelines
            pipelines = [
                pipeline_templates[num_stages]
                for num_stages in num_hosts_per_pipeline
                if num_stages > 0
            ]

            _, num_microbatches = pipeline_instantiator.distribute_batch(
                dict(Counter(pipelines)), need_all_pipelines_have_batch=True
            )
        else:
            num_instances, num_microbatches = pipeline_instantiator.instantiate(
                len(configuration_engine.dist_info)
            )

            pipelines = list(
                itertools.chain.from_iterable(
                    itertools.repeat(template, num_templates)
                    for template, num_templates in num_instances.items()
                )
            )

        return pipelines, num_microbatches

    @torch.no_grad()
    def reconfigure(
        self,
        pipeline_templates: dict[int, PipelineTemplate],
        model: HeterogeneousParallelModule,
        optimizer: HybridParallelAMPOptimizer | HybridParallelNaiveOptimizer,
        dataloader: HeterogeneousDataLoader,
        lr_scheduler: LRScheduler | None = None,
    ) -> tuple[
        HeterogeneousParallelPlugin,
        HybridParallelAMPOptimizer | HybridParallelNaiveOptimizer,
        HeterogeneousDataLoader,
        LRScheduler | None,
    ]:
        configuration_engine = ConfigurationEngine.get_instance()
        device = get_accelerator().get_current_device()

        # all layers in the model
        layers = list(
            itertools.chain.from_iterable(self.pipelines[0].modules_per_stage)
        )
        num_layers = len(layers)

        # Backup old attributes for reconfiguration before resetting
        old_pg_mesh: np.ndarray = copy.deepcopy(self.pg_mesh.mesh)
        old_rank_map = copy.deepcopy(configuration_engine.rank_map)

        old_my_layers = torch.tensor(
            [
                True
                if index in [coord[PP_AXIS] for coord in self.pg_mesh.coords]
                else False
                for index in range(num_layers)
            ],
            dtype=torch.bool,
            device=device,
        )

        # Reset process group
        configuration_engine.get_host_update()
        del self.pg_mesh
        self.pg_mesh = None
        configuration_engine.init_distributed()

        def all_gather_layers_per_rank(
            data: torch.Tensor, device: torch.device
        ) -> dict[tuple[int], np.ndarray]:
            data_list = torch.empty(
                configuration_engine.world_size,
                *data.shape,
                device=device,
                dtype=data.dtype,
            )
            dist.all_gather_into_tensor(data_list, data)

            data_list = data_list.numpy(force=True)
            result = {}
            for i in range(
                0,
                configuration_engine.world_size,
                self.shard_config.tensor_parallel_size,
            ):
                ranks = list(range(i, i + self.shard_config.tensor_parallel_size))
                assert all(
                    np.array_equal(data_list[i], data_list[i + tp_index])
                    for tp_index in range(self.shard_config.tensor_parallel_size)
                ), f"Data mismatch between ranks in the same tensor parallel group: {ranks}."
                result[tuple(ranks)] = data_list[i]

            return result

        # each tensor indicates which layers are held by each (new) rank
        old_layers_per_ranks = all_gather_layers_per_rank(old_my_layers, device)
        del old_my_layers

        # Prepare layer transfer before setting new pipelines
        # Setting new pipelines will free unused layers, which some other may need.

        # Create new pipelines.
        # TODO: If there is no feasible pipeline, merge pipelines
        new_pipelines, new_num_microbatches = self._instantiate_pipelines(
            pipeline_templates,
            self.global_batch_size // self.microbatch_size,
            old_pg_mesh,
            old_rank_map,
        )

        new_pg_mesh = HeterogeneousProcessGroupMesh(
            new_pipelines, self.shard_config.tensor_parallel_size
        )

        new_my_layers = torch.tensor(
            [
                True
                if index in [coord[PP_AXIS] for coord in new_pg_mesh.coords]
                else False
                for index in range(num_layers)
            ],
            dtype=torch.bool,
            device=device,
        )

        new_layers_per_ranks = all_gather_layers_per_rank(new_my_layers, device)

        """
        Missing layer transfer based on new pipeline configuration
        old_layers_per_dp_rank (numpy.ndarray[dp_size, num_layers]):
            For each dp rank, indicates which layers are held by the replica.
            True if the dp group holds the layer, False otherwise.
            dp_size is the size after failures which is newly assigned.
        new_layers_per_dp_rank (numpy.ndarray[dp_size, num_layers]):
            For each dp rank, indicates which layers should be held by the replica.
            True if the dp group should hold the layer, False if should not.

        1. Calculate required layers (layers that are now need but not held)
        2. For each required layer, find a rank that holds the layer
        3. Send the layer to the rank that needs the layer
        4. Update optimizer states and model parameters
        5. Calculate non-required layers (layers that are held but not needed)
        6. Free non-required layers
        """

        # Calculate required layers (layers that are now need but not held)
        # layer_index -> list of tuples of ranks (each tuple includes ranks per tensor parallel group)
        # Even if tensor parallelism is not used, it is a tuple with one element.
        layers_required_by_ranks: dict[int, list[tuple[int]]] = defaultdict(list)
        for ranks, has_layers in old_layers_per_ranks.items():
            for layer_index, has_layer in enumerate(has_layers):
                if not has_layer and new_layers_per_ranks[ranks][layer_index]:
                    layers_required_by_ranks[layer_index].append(ranks)

        layer_modules: dict[str, nn.Module] = {
            layer_name: module
            for name, module in model.module.named_modules()
            for layer_name in layers
            if name == layer_name
        }

        def get_layer_holder_ranks(layer_index: int) -> list[tuple[int]]:
            layer_holder_ranks = []
            for ranks, has_layers in old_layers_per_ranks.items():
                if has_layers[layer_index]:
                    layer_holder_ranks.append(ranks)
            return layer_holder_ranks

        adds_to_param_info: dict[str, dict] = {
            "param2id": {},
            "id2param": {},
            "param2shape": {},
        }
        removes_from_param_info: dict[str, list] = {
            "param2id": [],
            "id2param": [],
            "param2shape": [],
        }

        for layer_index, ranks in layers_required_by_ranks.items():
            layer_holder_ranks = get_layer_holder_ranks(layer_index)
            if not layer_holder_ranks:
                logger.error(
                    f"Currently no rank has layer: {layers[layer_index]}. "
                    "Please rerun training from last checkpoint."
                )
                raise RuntimeError(f"No one holds the layer: {layers[layer_index]}!")

            module = layer_modules[layers[layer_index]]

            for ranks_need_layer in ranks:
                if not layer_holder_ranks:
                    # Empty holder "here" means that all holders is consumed and has a job.
                    # Add more jobs to each holder.
                    layer_holder_ranks = get_layer_holder_ranks(layer_index)

                sender_ranks = layer_holder_ranks.pop(0)
                if configuration_engine.rank in ranks_need_layer:
                    # Find the corresponding tensor parallel rank in sender_ranks
                    sender_rank = sender_ranks[
                        ranks_need_layer.index(configuration_engine.rank)
                    ]
                    for (
                        submodule,
                        name,
                        placeholder,
                    ) in ModelSharder.buffer_placeholders(
                        module, delete_placeholders_after=True
                    ):
                        setattr(submodule, name, placeholder.create())

                    for (
                        submodule,
                        name,
                        placeholder,
                    ) in ModelSharder.parameter_placeholders(
                        module, delete_placeholders_after=True
                    ):
                        size: torch.Tensor = torch.empty(
                            1, dtype=torch.int64, device=device
                        )
                        dist.recv(size, sender_rank)

                        tensor: torch.Tensor = torch.empty(
                            size.item(), dtype=torch.uint8, device=device
                        )
                        dist.recv(tensor, sender_rank)
                        buff = io.BytesIO(tensor.numpy(force=True))
                        tensor: dict[str, tuple[dict, torch.Tensor]] = torch.load(
                            buff, map_location=device
                        )

                        states: dict[str, torch.Tensor] = tensor["states"]
                        master_tensor: torch.Tensor = tensor["parameter"]

                        # Check whether shape is correct
                        assert master_tensor.shape == placeholder.shape, (
                            f"{name} Shape mismatch between placeholder and parameter. "
                            f"received parameter shape: {master_tensor.shape}, "
                            f"placeholder shape: {placeholder.shape}"
                        )

                        p = torch.nn.Parameter(
                            master_tensor.to(dtype=model.mixed_precision)
                            if isinstance(optimizer, MixedPrecisionOptimizer)
                            and master_tensor.dtype == torch.float32
                            else master_tensor
                        )
                        setattr(submodule, name, p)

                        # TODO: check if it is still tensor (not nn.parameter) if we do fp32 training
                        optimizer.optim.state[master_tensor] = states
                        optimizer.optim.param_groups[0]["params"].append(master_tensor)

                        if isinstance(optimizer, MixedPrecisionOptimizer):
                            optimizer.master_to_working_map[master_tensor] = p
                            optimizer.working_to_master_map[p] = master_tensor

                        # Cache updates and apply them later
                        # This is because of potential duplicated IDs:
                        # We maintain IDs for parameters that have already been freed,
                        # some additional parameters may take the same ID.
                        # In this case sequentially updating param_info messes up.
                        optim_param_index = optimizer.param_info["param2id"][
                            placeholder.param_id
                        ]
                        removes_from_param_info["param2id"].append(placeholder.param_id)
                        removes_from_param_info["id2param"].append(optim_param_index)
                        removes_from_param_info["param2shape"].append(
                            placeholder.param_id
                        )

                        adds_to_param_info["param2id"][id(p)] = optim_param_index
                        adds_to_param_info["id2param"][optim_param_index] = id(p)
                        adds_to_param_info["param2shape"][id(p)] = p.shape

                elif configuration_engine.rank in sender_ranks:
                    # Find the corresponding tensor parallel rank in ranks_need_layer
                    rank_need_layer = ranks_need_layer[
                        sender_ranks.index(configuration_engine.rank)
                    ]
                    for name, param in module.named_parameters():
                        buff = io.BytesIO()
                        # Find master weights
                        master_param = (
                            optimizer.working_to_master_map[param]
                            if isinstance(optimizer, HybridParallelAMPOptimizer)
                            else param
                        )

                        try:
                            # Pickle optimizer state
                            states = optimizer.optim.state[master_param]
                        except KeyError:
                            states = None

                        torch.save(
                            {
                                "states": states,
                                "parameter": master_param,
                            },
                            buff,
                        )
                        buff.seek(0)

                        tensor: torch.Tensor = torch.frombuffer(
                            buff.getbuffer(), dtype=torch.uint8
                        ).to(device)
                        size = torch.tensor(
                            [tensor.numel()], dtype=torch.int64, device=device
                        )

                        dist.send(size, rank_need_layer)
                        dist.send(tensor, rank_need_layer)

        # Update internal process group in ParallelModules
        my_layers = [
            layers[index] for index, has_layer in enumerate(new_my_layers) if has_layer
        ]
        tp_group = new_pg_mesh.get_group_along_axis(TP_AXIS)

        for layer in my_layers:
            module = layer_modules[layer]
            # Replace process group in ParallelModules

            for submodule in module.modules():
                if isinstance(submodule, ParallelModule):
                    for attr_name in dir(submodule):
                        attr = getattr(submodule, attr_name)
                        if isinstance(attr, dist.ProcessGroup):
                            logger.debug(
                                f"Replacing process group {attr} with {tp_group}"
                            )
                            setattr(submodule, attr_name, tp_group)

        for param_info_key, items in removes_from_param_info.items():
            for item in items:
                del optimizer.param_info[param_info_key][item]

        for param_info_key, items in adds_to_param_info.items():
            param_info: dict = optimizer.param_info[param_info_key]
            param_info.update(items)

        # free no-longer necessary parameter here
        # TODO: must create _buffer_placeholders and _parameter_placeholders
        old_layers = next(
            layers
            for ranks, layers in old_layers_per_ranks.items()
            if configuration_engine.rank in ranks
        )
        new_layers = next(
            layers
            for ranks, layers in new_layers_per_ranks.items()
            if configuration_engine.rank in ranks
        )
        for index, (old_layer, new_layer) in enumerate(zip(old_layers, new_layers)):
            if old_layer and not new_layer:
                module = layer_modules[layers[index]]
                ModelSharder.set_tensors_to_placeholder(module)

        self.set_pipelines(new_pipelines, new_num_microbatches)
        model, optimizer, _, dataloader, lr_scheduler = self.configure(
            model, optimizer, None, dataloader, lr_scheduler, forced=True
        )

        return model, optimizer, dataloader, lr_scheduler



from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class HostInfo(_message.Message):
    __slots__ = ("ip", "devices", "port", "status")
    IP_FIELD_NUMBER: _ClassVar[int]
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ip: str
    devices: str
    port: int
    status: str
    def __init__(self, ip: _Optional[str] = ..., devices: _Optional[str] = ..., port: _Optional[int] = ..., status: _Optional[str] = ...) -> None: ...

class DistInfo(_message.Message):
    __slots__ = ("hosts",)
    HOSTS_FIELD_NUMBER: _ClassVar[int]
    hosts: _containers.RepeatedCompositeFieldContainer[HostInfo]
    def __init__(self, hosts: _Optional[_Iterable[_Union[HostInfo, _Mapping]]] = ...) -> None: ...

class CodeInfo(_message.Message):
    __slots__ = ("path", "args")
    PATH_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    path: str
    args: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, path: _Optional[str] = ..., args: _Optional[_Iterable[str]] = ...) -> None: ...

class PortInfo(_message.Message):
    __slots__ = ("port",)
    PORT_FIELD_NUMBER: _ClassVar[int]
    port: int
    def __init__(self, port: _Optional[int] = ...) -> None: ...

class AgentInfo(_message.Message):
    __slots__ = ("agent_index",)
    AGENT_INDEX_FIELD_NUMBER: _ClassVar[int]
    agent_index: int
    def __init__(self, agent_index: _Optional[int] = ...) -> None: ...



from __future__ import annotations

import multiprocessing
import socket
import sys
from concurrent import futures
from concurrent.futures import Future, ProcessPoolExecutor, ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from multiprocessing.context import SpawnProcess
from multiprocessing.synchronize import Condition
from pathlib import Path

import click
import fabric
import grpc
from google.protobuf import empty_pb2
from loguru import logger

from oobleck.elastic import master_service_pb2, master_service_pb2_grpc

"""
Oobleck master process code.
The master transfers the given serialized execution code
to agents, which will be executed in their worker processes.

After launching all agent processes, the master process
watches disconnection evernts from agents.
Once an agent is disconnected, the master process will
broadcast `reconfigure` message to all live agents.
"""

agent_list: list[tuple[HostInfo, Future]] = []


@dataclass
class LaunchArguments:
    # Path to the hostfile
    hostfile: Path
    # A tag to identify this run
    tag: str
    # Port for master gRPC service
    master_service_port: int
    # Oobleck root directory to store logs and profiles.
    base_dir: Path
    # Print agent's ssh outputs to stdout, instead of files
    debug: bool


@dataclass
class ScriptArguments:
    training_script: Path
    training_script_args: list[str]


class HostStatus(Enum):
    up = 0
    killed = 1
    terminating = 2


@dataclass
class HostInfo:
    ip: str
    devices: str
    port: int
    status: HostStatus = HostStatus.up

    def __eq__(self, other: HostInfo) -> bool:
        return (
            self.ip == other.ip
            and self.devices == other.devices
            and self.port == other.port
        )

    def __hash__(self) -> int:
        return hash((self.ip, self.devices, self.port))

    @staticmethod
    def fetch_hostfile(hostfile_path: Path) -> list[HostInfo]:
        """
        Parse the hostfile (MPI style) and return a list of HostInfo objects.

        A hostfile should look like:
        worker-0 slots=2 devices=0,1 port=22
        worker-0 slots=2 devices=2,3 port=22
        worker-1 slots=2 port=1234
        worker-1 slots=2 port=1235

        The `devices` and `port` fields are optional.

        You must specify the same number of slots to all agents.

        You can optionally specify the `devices` field to specify which GPUs
        should be used by the agent on the host.
        This will allow you to run multiple agents on the same host, with different GPUs.
        If not specified, `slots` number of GPUs starting from 0 will be used.

        If you use Docker containers to split GPUs on the same host,
        you can specify different port numbers for each container.
        """
        hosts: list[HostInfo] = []
        with hostfile_path.open("r") as f:
            for line in f.readlines():
                parts = line.split()
                # skip empty lines
                if not parts:
                    continue

                ip, slots, devices, port = (
                    socket.gethostbyname(parts[0]),
                    None,
                    None,
                    None,
                )
                first_slots = None
                for part in parts[1:]:
                    if part.startswith("slots="):
                        slots = int(part.split("=")[1])

                        if first_slots is None:
                            first_slots = slots

                        if first_slots != slots:
                            raise ValueError(
                                "All agents must have the same number of slots."
                            )
                    elif part.startswith("devices="):
                        devices = part.split("=")[1]
                    elif part.startswith("port="):
                        port = int(part.split("=")[1])

                if slots is None:
                    raise ValueError(
                        "The `slots` field must be specified for every agent."
                    )

                if devices is None:
                    devices = ",".join(str(i) for i in range(slots))
                if port is None:
                    port = 22

                host_info = HostInfo(ip, devices, port)
                if any(host == host_info for host in hosts):
                    raise ValueError(f"Duplicated host: {host_info}")

                hosts.append(host_info)

        logger.debug(f"Hosts: {hosts}")

        return hosts


class MultiNodeAgentRunner:
    """
    A runner to execute multiple agents on multiple nodes.
    """

    def __init__(
        self,
        disconnect_condition: Condition,
        hosts: list[HostInfo],
        master_service_port: int,
        tag: str,
        base_dir: Path | None = None,
    ):
        self.disconnect_condition = disconnect_condition
        self.hosts = hosts
        self.master_service_port = master_service_port
        self.tag = tag
        self.base_dir = base_dir

    @staticmethod
    def run_on_nodes(
        agent_index: int,
        host: HostInfo,
        master_service_port: int,
        tag: str,
        base_dir: Path,
        debug: bool,
    ):
        """
        Use fabric to run the agent on the given host.
        This function will block until the agent process is terminated.
        Therefore, it must be executed on a separate process.
        """
        my_ip = socket.gethostbyname(socket.gethostname())

        logger.info(
            f"Connecting to {host.ip}:{host.port} to instantiate an agent {agent_index}..."
        )

        try:
            with fabric.Connection(host.ip, port=host.port, connect_timeout=30) as conn:
                cmd = f"{sys.executable} -m oobleck.elastic.agent "
                cmd += f"--master_ip {my_ip} --master_port {master_service_port} "
                cmd += f"--agent_index {agent_index} "
                cmd += f"--tag {tag} --base_dir {str(base_dir)}"

                logger.debug(f"Connected to {host.ip}:{host.port}. Executing: {cmd}")

                if not debug:
                    out_stream = (base_dir / tag / f"agent{agent_index}.log").open("w")
                    logger.info(
                        f"Agent {agent_index} output will be saved to {(out_stream.name)}."
                    )
                else:
                    out_stream = sys.stderr

                conn.run(
                    cmd,
                    hide=True,
                    out_stream=out_stream,
                    err_stream=out_stream,
                )

        except Exception as e:
            logger.warning(f"[Agent {agent_index}] SSH disconnected: {e}")
            raise

        logger.info(f"[Agent {agent_index}] done. Exiting...")

    def run(self, debug: bool = False) -> list[SpawnProcess]:
        """
        Spawn multiple processes to run agents on multiple hosts.
        Each process accesses a host via SSH and runs the agent.
        """
        context = multiprocessing.get_context("spawn")
        with ProcessPoolExecutor(
            max_workers=len(self.hosts), mp_context=context
        ) as executor:
            global agent_list
            for agent_index, host in enumerate(self.hosts):
                future = executor.submit(
                    self.run_on_nodes,
                    agent_index,
                    host,
                    self.master_service_port,
                    self.tag,
                    self.base_dir,
                    debug,
                )
                agent_list.append((host, future))

            # Loop until all processes done
            while agent_list:
                not_done: list[Future] = [f for _, f in agent_list]
                done, not_done = futures.wait(
                    not_done, return_when=futures.FIRST_EXCEPTION
                )

                reconfiguration_needed = False
                for f in done:
                    if f.exception() is not None:
                        reconfiguration_needed = True

                    value = next(
                        (host, future) for host, future in agent_list if future == f
                    )
                    agent_list.remove(value)

                logger.debug("Remaining agents: {}", agent_list)

                if reconfiguration_needed:
                    with self.disconnect_condition:
                        self.disconnect_condition.notify_all()

            assert not agent_list and not not_done, "All agents should be done by now."


class MasterService(master_service_pb2_grpc.OobleckMasterServicer):
    """
    Master gRPC service.

    This service is used to transfer the serialized distributed info and
    user code to agents. It also broadcasts agent disconnect events to
    all agents. Broadcasting is done with server to client streaming gRPC
    and conditional variable; the cv is rung when an agent disconnects
    by MultiNodeAgentRunner after a ssh session to an agent is disconnected.
    """

    def __init__(
        self,
        script_args: ScriptArguments,
        disconnect_condition: Condition,
    ):
        self.script_args = script_args
        self.disconnect_condition = disconnect_condition
        self.master_port = 0

    def GetDistInfo(
        self,
        request: master_service_pb2.DistInfo,
        context: grpc.RpcContext,
    ) -> master_service_pb2.DistInfo:
        return master_service_pb2.DistInfo(
            hosts=[
                master_service_pb2.HostInfo(
                    ip=host.ip,
                    devices=host.devices,
                    port=host.port,
                    status=host.status.name,
                )
                for host, _ in agent_list
            ]
        )

    def GetCode(
        self,
        request: master_service_pb2.CodeInfo,
        context: grpc.RpcContext,
    ) -> master_service_pb2.CodeInfo:
        return master_service_pb2.CodeInfo(
            path=self.script_args.training_script.absolute().as_posix(),
            args=self.script_args.training_script_args,
        )

    def SetMasterRankPort(
        self,
        request: master_service_pb2.PortInfo,
        context: grpc.RpcContext,
    ) -> empty_pb2.Empty:
        self.master_port = request.port
        return empty_pb2.Empty()

    def GetMasterRankPort(
        self,
        request: empty_pb2.Empty,
        context: grpc.RpcContext,
    ) -> master_service_pb2.PortInfo:
        return master_service_pb2.PortInfo(port=self.master_port)

    def KillAgent(
        self, request: master_service_pb2.AgentInfo, context: grpc.RpcContext
    ) -> empty_pb2.Empty:
        agent_index = request.agent_index
        host, _ = agent_list[agent_index]

        host.status = HostStatus.terminating
        logger.info(f"Terminating agent {agent_index} on {host.ip}:{host.port}")

        with self.disconnect_condition:
            self.disconnect_condition.notify_all()

        return empty_pb2.Empty()

    def WatchReconfigurationNotification(
        self,
        request: empty_pb2.Empty,
        context: grpc.RpcContext,
    ):
        with self.disconnect_condition:
            self.disconnect_condition.wait()

        if context.is_active():
            yield master_service_pb2.DistInfo(
                hosts=[
                    master_service_pb2.HostInfo(
                        ip=host.ip,
                        devices=host.devices,
                        port=host.port,
                        status=host.status.name,
                    )
                    for host, _ in agent_list
                ]
            )


@click.command(context_settings={"ignore_unknown_options": True})
@click.option("--hostfile", type=Path, help="Path to the hostfile.")
@click.option("--tag", type=str, help="A tag to identify this run.")
@click.option(
    "--master_service_port", type=int, default=0, help="Port for master gRPC service."
)
@click.option(
    "--base_dir",
    type=Path,
    help="Oobleck root directory store logs and profiles.",
    default=Path("/tmp/oobleck"),
)
@click.option(
    "--debug",
    type=bool,
    is_flag=True,
    help="Print agent's ssh outputs to stderr, instead of files.",
    default=False,
)
@click.argument("training_script", type=Path)
@click.argument("training_script_args", nargs=-1, type=click.UNPROCESSED)
def serve(
    hostfile: Path,
    tag: str,
    master_service_port: int,
    base_dir: Path,
    debug: bool,
    training_script: Path,
    training_script_args: list[str],
):
    """
    training_script: Full path to the training script to be launched in parallel,
    followed by all the arguments for the training script.
    """
    launch_args = LaunchArguments(
        hostfile=hostfile,
        tag=tag,
        master_service_port=master_service_port,
        base_dir=base_dir,
        debug=debug,
    )
    script_args = ScriptArguments(
        training_script=training_script,
        training_script_args=training_script_args,
    )

    logger.info(f"Dist arguments: {launch_args}")
    logger.info(f"Script arguments: {script_args}")

    job_directory = launch_args.base_dir / launch_args.tag
    job_directory.mkdir(parents=True, exist_ok=True)
    assert job_directory.is_dir(), f"{str(job_directory)} is not a directory."

    hostinfo = HostInfo.fetch_hostfile(launch_args.hostfile)

    server = grpc.server(ThreadPoolExecutor(max_workers=None))
    disconnect_condition = multiprocessing.get_context("spawn").Condition()
    service = MasterService(script_args, disconnect_condition)
    master_service_pb2_grpc.add_OobleckMasterServicer_to_server(service, server)
    port = server.add_insecure_port(f"0.0.0.0:{launch_args.master_service_port}")
    server.start()
    logger.info(f"Running master service on port {port}")

    runner = MultiNodeAgentRunner(
        disconnect_condition, hostinfo, port, launch_args.tag, launch_args.base_dir
    )
    runner.run(launch_args.debug)

    logger.info("Training is done. Stopping the master service.")
    server.stop(grace=None)


if __name__ == "__main__":
    serve()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

import oobleck.elastic.master_service_pb2 as master__service__pb2


class OobleckMasterStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetDistInfo = channel.unary_unary(
            "/OobleckMaster/GetDistInfo",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=master__service__pb2.DistInfo.FromString,
        )
        self.GetCode = channel.unary_unary(
            "/OobleckMaster/GetCode",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=master__service__pb2.CodeInfo.FromString,
        )
        self.SetMasterRankPort = channel.unary_unary(
            "/OobleckMaster/SetMasterRankPort",
            request_serializer=master__service__pb2.PortInfo.SerializeToString,
            response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
        self.GetMasterRankPort = channel.unary_unary(
            "/OobleckMaster/GetMasterRankPort",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=master__service__pb2.PortInfo.FromString,
        )
        self.WatchReconfigurationNotification = channel.unary_stream(
            "/OobleckMaster/WatchReconfigurationNotification",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=master__service__pb2.DistInfo.FromString,
        )
        self.KillAgent = channel.unary_unary(
            "/OobleckMaster/KillAgent",
            request_serializer=master__service__pb2.AgentInfo.SerializeToString,
            response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class OobleckMasterServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetDistInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetCode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SetMasterRankPort(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetMasterRankPort(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def WatchReconfigurationNotification(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def KillAgent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_OobleckMasterServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetDistInfo": grpc.unary_unary_rpc_method_handler(
            servicer.GetDistInfo,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=master__service__pb2.DistInfo.SerializeToString,
        ),
        "GetCode": grpc.unary_unary_rpc_method_handler(
            servicer.GetCode,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=master__service__pb2.CodeInfo.SerializeToString,
        ),
        "SetMasterRankPort": grpc.unary_unary_rpc_method_handler(
            servicer.SetMasterRankPort,
            request_deserializer=master__service__pb2.PortInfo.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
        "GetMasterRankPort": grpc.unary_unary_rpc_method_handler(
            servicer.GetMasterRankPort,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=master__service__pb2.PortInfo.SerializeToString,
        ),
        "WatchReconfigurationNotification": grpc.unary_stream_rpc_method_handler(
            servicer.WatchReconfigurationNotification,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=master__service__pb2.DistInfo.SerializeToString,
        ),
        "KillAgent": grpc.unary_unary_rpc_method_handler(
            servicer.KillAgent,
            request_deserializer=master__service__pb2.AgentInfo.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "OobleckMaster", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class OobleckMaster(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetDistInfo(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/OobleckMaster/GetDistInfo",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            master__service__pb2.DistInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetCode(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/OobleckMaster/GetCode",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            master__service__pb2.CodeInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SetMasterRankPort(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/OobleckMaster/SetMasterRankPort",
            master__service__pb2.PortInfo.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetMasterRankPort(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/OobleckMaster/GetMasterRankPort",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            master__service__pb2.PortInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def WatchReconfigurationNotification(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/OobleckMaster/WatchReconfigurationNotification",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            master__service__pb2.DistInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def KillAgent(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/OobleckMaster/KillAgent",
            master__service__pb2.AgentInfo.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: master_service.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14master_service.proto\x1a\x1bgoogle/protobuf/empty.proto\"E\n\x08HostInfo\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x65vices\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\r\x12\x0e\n\x06status\x18\x04 \x01(\t\"$\n\x08\x44istInfo\x12\x18\n\x05hosts\x18\x01 \x03(\x0b\x32\t.HostInfo\"&\n\x08\x43odeInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\t\"\x18\n\x08PortInfo\x12\x0c\n\x04port\x18\x01 \x01(\r\" \n\tAgentInfo\x12\x13\n\x0b\x61gent_index\x18\x01 \x01(\r2\xe5\x02\n\rOobleckMaster\x12\x32\n\x0bGetDistInfo\x12\x16.google.protobuf.Empty\x1a\t.DistInfo\"\x00\x12.\n\x07GetCode\x12\x16.google.protobuf.Empty\x1a\t.CodeInfo\"\x00\x12\x38\n\x11SetMasterRankPort\x12\t.PortInfo\x1a\x16.google.protobuf.Empty\"\x00\x12\x38\n\x11GetMasterRankPort\x12\x16.google.protobuf.Empty\x1a\t.PortInfo\"\x00\x12I\n WatchReconfigurationNotification\x12\x16.google.protobuf.Empty\x1a\t.DistInfo\"\x00\x30\x01\x12\x31\n\tKillAgent\x12\n.AgentInfo\x1a\x16.google.protobuf.Empty\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'master_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_HOSTINFO']._serialized_start=53
  _globals['_HOSTINFO']._serialized_end=122
  _globals['_DISTINFO']._serialized_start=124
  _globals['_DISTINFO']._serialized_end=160
  _globals['_CODEINFO']._serialized_start=162
  _globals['_CODEINFO']._serialized_end=200
  _globals['_PORTINFO']._serialized_start=202
  _globals['_PORTINFO']._serialized_end=226
  _globals['_AGENTINFO']._serialized_start=228
  _globals['_AGENTINFO']._serialized_end=260
  _globals['_OOBLECKMASTER']._serialized_start=263
  _globals['_OOBLECKMASTER']._serialized_end=620
# @@protoc_insertion_point(module_scope)
